# built-in py modules
import json

# 3rd party modules
import dash
from dash.dependencies import Input, Output, State
from dash.exceptions import PreventUpdate

# Local modules
from app import app

@app.callback(Output('graph', 'figure'),
              [Input('fresnel_slider','value')])
def change_lighting(fresnel):
    return fresnel

#@app.callback(Output('hover-data', 'children'),
#              [Input('graph', 'hoverData')])

#def display_hover_data(hover_data):
#    return json.dumps(hover_data, indent=2)

#@app.callback(Output('hover-data', 'children'),
#              [Input('graph', 'figure')],
#                [State('graph', 'figure')])

#def display_hover_data(figure, statefig):
#    return json.dumps(statefig['layout']['scene'], indent=2)



import pathlib

DATA_PATH = pathlib.Path.cwd().joinpath('data').resolve()
GEOMETRY_DIR = 'obj' #(in DATA_PATH)
RESULTS_DIR = 'results' #(in DATA_PATH)

# found this somewhere, thought it might me useful.
COLOR_SCALES  = ['Blackbody',
'Bluered',
'Blues',
'Earth',
'Electric',
'Greens',
'Greys',
'Hot',
'Jet',
'Picnic',
'Portland',
'Rainbow',
'RdBu',
'Reds',
'Viridis',
'YlGnBu',
'YlOrRd']

SCHEME_PARAMETER = 'scheme'

SCENARIO_PARAMETERS = [SCHEME_PARAMETER, 'roof_area', 'roof_vlt' ,'fac_vlt', 'can_vlt']

SCHEMES = {
    'tri' : 'Triangular',
    'current' : 'PD radial strips',
    'strip' : 'Circumferential strips'
    }



#built-in py modules
import os

#3rd party modules
import dash
import dash_core_components as dcc
import dash_html_components as html
import dash_colorscales as dcs
import dash_bootstrap_components as dbc
from dash.exceptions import PreventUpdate


# local modules
import geometry_tools

axis_template = {
    "showbackground": False,
    "visible" : False
}

plot_layout = {
    "title": "",
    "margin": {"t": 0, "b": 0, "l": 0, "r": 0},
    "font": {"size": 12, "color": "white"},
    "showlegend": False,
    'uirevision':'same_all_the_time', #this keeps camera position etc the same when data changes.
    "scene": {
        "xaxis": axis_template,
        "yaxis": axis_template,
        "zaxis": axis_template,
        "aspectmode" : "data",
        "camera": {"eye": {"x": 1.25, "y": 1.25, "z": 1.25}},
        "annotations": [],
    },
}


control_panel = [
    html.Pre(
        id = 'hover-data',
    ),
    dcc.Slider(
        id = 'fresnel_slider',
        min=0,
        max=10,
        step=0.5,
        value=1
    )  
]

# control box type overlay, on left of screen
overlay = dbc.Container(
            dbc.Row([
                dbc.Col(
                    children = dbc.FormGroup(control_panel)
                , width=2),
            ],justify='start'
        ),id = 'overlay', fluid = True
    )


def make_layout():
    """

    """
    return html.Div([dcc.Graph(
                        id="graph",
                        figure={
                            "data": geometry_tools.import_geometry(["test"]),
                            "layout": plot_layout,
                        },
                        config={"scrollZoom": True}, # activates wheel thingy on mouse to zoom and wotnot
                    ),
                     overlay
                     ])






# builtin modules
import pathlib
from typing import List

#3rd party modules
import numpy as np
import plotly.graph_objs as go

#Local modules
import _config 
import wav_obj

def make_ployly_mesh3d(
    vertices,
    faces,
    name = '',
    color='grey',
    colorscale = None,
    opacity = 1,
    flatshading=False,
    intensities = None,
    facecolors = None,
    showscale=False,
    hoverinfo='none',
    hovertemplate = '',
    customdata = None
):

    x, y, z = vertices.T
    I, J, K = faces.T
  

    trace = go.Mesh3d(
        x=x,
        y=y,
        z=z,
        colorscale = colorscale,
        color = color,
        intensity = intensities,
        facecolor= facecolors,
        opacity = opacity,
        flatshading = flatshading,
        hoverinfo=hoverinfo,
        hovertemplate = hovertemplate,
        i = I,
        j = J,
        k = K,
        customdata = None,
        name = name,
        showscale = showscale,
        lighting = {
            "ambient": 0.0,
            "diffuse": 0.8,
            "fresnel": 1,
            "specular": 1.5,
            "roughness": 1,
        },
        lightposition = {"x": 1000, "y": 2000, "z": 1000},   
        )

    return trace




def import_geometry(obj_names : List[str], path = _config.GEOMETRY_DIR):
    c = []
    for component in obj_names:
        traces = create_mesh_data(component)
        c.extend(traces)
    return c


def create_mesh_data(component : str, path = _config.GEOMETRY_DIR):
    """
    Create mesh-for-plotly from single obj file
    """

    data = []
    wav_objs = wav_obj.WavObject.read_objfile(component, path, split = False)

    for i, obj in enumerate(wav_objs):
        vertices = obj.vertices
        faces = obj.tri_faces
        facecolors = obj.face_mtls
        mesh = make_ployly_mesh3d(
            obj.vertices,
            obj.tri_faces,
            name = obj.name,
            facecolors = obj.face_colors,
            opacity = obj.opacities[0]
        )
        data.append(mesh)

    return data




# built-in py modules
from typing import List, Iterator

import pathlib

# 3rd party modules
from multimethod import multimeta
import numpy as np


# Local modules
import _config 

class WavObject(metaclass=multimeta):
    """
    Parse 1 or many objects in wavefront obj file using one or other of static methof:
        WavObject.read_objfile(...)
    """

    COMMENT_LINE = '#' #this and everything else except the above is currently ignored.

    VERTEX_LINE = 'v '
    FACE_LINE = 'f '
    OBJECT_NAME_LINE = 'o '
    MATERIAL_LIB_LINE = 'mtllib '
    USE_MATERIAL_LINE = 'usemtl '

    def __init__(self):
        self.name : str = None 
        self.vertices  = []
        self.tri_faces = []
        self.face_mtls: List(str) = []
        self.face_colors: List(List(int)) = []
        self.opacities: List(float) = []

   
    def read_objfile(full_obj_filepath : pathlib.Path, split = True):
        """
        NOTE : This is a @staticmethod of WavObject, but using @staticmethod decorator 
        seem to break multimethod

        Parses a wavefront obj file.
    
        Args:
            full_obj_filepath (str) : full os-specific filepath of .obj file (e.g. monkey.obj)
            split (bool) : if True then returns all objects in file as separate elements
        Returns: 
            my_objects : List of WavObjects
        """
        my_objects = [] # can be multiple sub objects in file,if split = True


        def triangulate_polygons(list_vertex_indices):
            for k in range(0, len(list_vertex_indices), 3):
                yield list_vertex_indices[k : k + 3]


        def process_faces(current_obj,face_mtls,index_correction, correction = 1):
            """
            correction = 1 for dash. 
            """
            current_obj.vertices = np.array(current_obj.vertices)
            current_obj.tri_faces = [int(i) for i in current_obj.tri_faces]
            current_obj.tri_faces = np.array(list(triangulate_polygons(current_obj.tri_faces))) - (correction + index_correction- (0 if len(my_objects) == 0 else 1))
            
            # currently we discard everything except the kd (diffuse color) 
            current_obj.face_colors = list(map(lambda x : x.kd,face_mtls))
            current_obj.opacities = list(map(lambda x : 1-x.tr,face_mtls))
            my_objects.append(current_obj)


        obj_mtls = {}
        vtx_indices = []

        # face_mtls this is just a temp array that is 'processed' in process_faces function
        # then cleared for next object
        face_mtls = [] 
        current_mtl = _BasicWavMaterial()
        with open(full_obj_filepath) as fp:
            index_correction = 0

            count=0 # running count of vertices in whole file
            prev_good_line = WavObject.VERTEX_LINE 
            current_obj = WavObject() # first object
            for i, line in enumerate(fp):
                
                if line.startswith(WavObject.VERTEX_LINE):
                    count = count+1
                    if prev_good_line == WavObject.FACE_LINE and split:
                        # new mesh object in file
                        # process & push current one into our my_objects list
                        process_faces(current_obj,face_mtls,index_correction)
                        
                        index_correction = count
                        #and start again
                        current_obj = WavObject()
                        face_mtls = []

                    line = line[2:].strip() # remove v and whitespace
                    current_obj.vertices.append(list(map(float, line.split())))
                    prev_good_line = WavObject.VERTEX_LINE
                elif line.startswith(WavObject.FACE_LINE):
                    line = line[len(WavObject.FACE_LINE):].strip()
                    vtx_indices = [x.split('/')[0] for x in line.split()]
                    if len(vtx_indices) != 3:
                        raise NotImplementedError("Currently all faces have to be triangular")
                    face_mtls.append(current_mtl)
                    current_obj.tri_faces.extend(vtx_indices)
                    prev_good_line = WavObject.FACE_LINE
                elif line.startswith(WavObject.OBJECT_NAME_LINE):
                    line = line[len(WavObject.OBJECT_NAME_LINE):].strip()
                    current_obj.name = line
                    #todo this is a bit crude. Might not work 100% of the time. 
                elif line.startswith(WavObject.MATERIAL_LIB_LINE):
                    mtl_file = line[len(WavObject.MATERIAL_LIB_LINE):].strip() # remove mtllib and whitespace
                    full_mtl_filepath = full_obj_filepath.parents[0].joinpath(mtl_file)
                    obj_mtls.update(_BasicWavMaterial.read_mtlfile(full_mtl_filepath))
                elif line.startswith(WavObject.USE_MATERIAL_LINE):
                    mtl_name = line[len(WavObject.USE_MATERIAL_LINE):].strip() # remove usemtl and whitespace
                    if mtl_name in obj_mtls:
                        current_mtl = obj_mtls[mtl_name]



        process_faces(current_obj,face_mtls, index_correction) # add last object (or only object!)
        return my_objects


    def read_objfile(component_name : str, sub_dir : str , split = True):
        """
        NOTE : This is a @staticmethod of WavObject, but using @staticmethod decorator 
        seem to break multimethod

        Overloaded method for loading obj files.
        """
        if sub_dir is not None:
            full_filepath = _config.DATA_PATH.joinpath(sub_dir, component_name + ".obj")
        else:
            full_filepath = _config.DATA_PATH.joinpath(component_name + ".obj")

        return WavObject.read_objfile(full_filepath,split)




###############################################################################################################

class _BasicWavMaterial:
    """
    Basic material 
    https://en.wikipedia.org/wiki/Wavefront_.obj_file#Basic_materials
    """

    NEW_MATERIAL = "newmtl "
    AMBIENT_COLOR = "Ka " # e.g. 1.000 1.000 1.000   is white
    DIFFUSE_COLOR = "Kd "
    SPECULAR_COLOR = "Ks "
    TRANSPARENCY = "Tr "
    DISSOLVE = "d "
    TRANSMISSION = "Tf "
    SPECULAR_EXPONENT = "Ns "


    def __init__(self, name = "default_blue"):
        """

        """
        self.name :str = name 
        self.ka : List(int)= [0.0,0.0,0.0] #ambient color. R G B
        self.kd : List(int) = [0.0,0.0,255]  #diffuse color. R G B We default to blue
        self.ks : List(int) = [255,255,255] #spec color
        self.tr : float = 0.0 #transparency: sometime d (dissolve) = 1.0 - Tr
        self.tf : List(int) = [0,0,0] #transmission R G B
        self.Ns : float = 0.0 #specular exponent...the focus of specular highlight. typ 0 to 1000 

    def __str__(self):
        return " Basic material: %s ka: %s kd : %s  ks %s" % (self.name, self.ka,self.kd,self.ks)

    def __repr__(self):
        return self.__str__()

    @staticmethod
    def read_mtlfile(full_mtl_filepath : pathlib.Path):


        def rgb(line : str):
            """
            converts obj-format indexed color string to rgb integer array
            eg 0 0.5 1.0 -> [0 128 255]
            """
            return [int(i* 255) for i in map(float, line.strip().split())]

        current_mtl = None        
        my_mtls = {} 

        with open(full_mtl_filepath) as fp:
            for i, line in enumerate(fp):
                if line.startswith(_BasicWavMaterial.NEW_MATERIAL):
                    if current_mtl is not None:
                        my_mtls[current_mtl.name] = current_mtl
                    # start new material
                    name = line[len(_BasicWavMaterial.NEW_MATERIAL):].strip()
                    current_mtl = _BasicWavMaterial(name)
                elif line.startswith(_BasicWavMaterial.AMBIENT_COLOR):
                    line = line[len(_BasicWavMaterial.AMBIENT_COLOR):] # remove ka and whitespace
                    current_mtl.ka = rgb(line)
                elif line.startswith(_BasicWavMaterial.DIFFUSE_COLOR):
                    line = line[len(_BasicWavMaterial.DIFFUSE_COLOR):] # remove kd and whitespace
                    current_mtl.kd = rgb(line)
                elif line.startswith(_BasicWavMaterial.SPECULAR_COLOR):
                    line = line[len(_BasicWavMaterial.SPECULAR_COLOR):] # remove ks and whitespace
                    current_mtl.ks = rgb(line)
                elif line.startswith(_BasicWavMaterial.TRANSPARENCY):
                    line = line[len(_BasicWavMaterial.TRANSPARENCY):].strip() # remove tr and whitespace
                    current_mtl.tr = float(line)
                elif line.startswith(_BasicWavMaterial.DISSOLVE):
                    line = line[len(_BasicWavMaterial.DISSOLVE):].strip() # remove d and whitespace
                    current_mtl.tr = 1.0 - float(line)
                elif line.startswith(_BasicWavMaterial.TRANSMISSION):
                    line = line[len(_BasicWavMaterial.TRANSMISSION):] # remove tf and whitespace
                    current_mtl.tr = rgb(line)
        
        #pick up the last one.                       
        if current_mtl is not None:
            my_mtls[current_mtl.name] = current_mtl

        return my_mtls



"""A setuptools based setup module.

See:
https://packaging.python.org/guides/distributing-packages-using-setuptools/
https://github.com/pypa/sampleproject
"""

from setuptools import setup, find_packages
from os import path

HERE = path.dirname(path.abspath(__file__))

GITHUB_URL = "https://github.com/aegis1980/dash-obj-in-3dmesh/"

def _get_long_description():
    """
    Get the long description from the README file
    """
    with open(path.join(HERE, "README.md")) as f:
        return f.read()


setup(
    name='dash-obj-in-3dmesh',
    version='0.3.1',
    description='Some tools for getting Wavefront OBJ files into a Plotly Dash 3dmesh graph',
    long_description=_get_long_description(),
    long_description_content_type='text/markdown', 
    author='Jon Robinson',
    author_email='jonrobinson1980@gmail.com',
    license='CC Attribution 4.0 International',
    packages=find_packages(),
    python_requires='>=3.5',
    install_requires=[
        'numpy', 
        'dash',
        'multimethod'
    ], #external packages as dependencies
    classifiers=[
        "Framework :: Dash"
    ]
)



import dash
import dash_bootstrap_components as dbc


app = dash.Dash(
    __name__,
    external_stylesheets=[dbc.themes.LUX]
 )

# flask server (parent)
server = app.server


# Since we're adding callbacks to elements that don't exist in the app.layout,
# Dash will raise an exception to warn us that we might be
# doing something wrong.
# In this case, we're adding the elements through a callback, so we can ignore
# the exception.
app.config.suppress_callback_exceptions = True